/************************************************************************
* "Epsilon", "EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd. *
*                                                                       *
*  Copyright (C) 1985, 1990 Lugaru Software Ltd.  All rights reserved.  *
*                                                                       *
* Limited permission is hereby granted to reproduce and modify this     *
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in  *
* any such reproduction or modification.                                *
************************************************************************/

#include "eel.h"
#include "codes.h"

/************************************************************************
*                                                                       *
*   command  copy_entire_line()                                         *
*   command  copy_entire_word()                                         *
*   command  copy_mark()                                                *
*   command  kill_entire_line()                                         *
*   command  kill_entire_word()                                         *
*   command  kill_mark()                                                *
*   command  set_mark_block()                                           *
*   command  set_mark_lines()                                           *
*   command  set_mark_region()                                          *
*   command  yank_mark()                                                *
*   command  yank_pop_mark()                                            *
*                                                                       *
************************************************************************/


#define  SEARCH_FORWARD     1
#define  SEARCH_BACKWARD   -1

#define  COPY_BLOCK         'c'
#define  KILL_BLOCK         'k'
#define  DELETE_BLOCK       'd'

buffer char _is_line = 0;
int _oldfrom = -1, _oldto = -1;	/* region yanked last time */

#define UNDO_FLAG_KILL	1	/* bracket each yank in undo with these */


command copy_mark()
  {
    int high;
        
    if (in_rectangle_mode)
      {
        copy_region();
        return;
      }
    
    high = is_highlight_on();
    highlight_off();
    
    switch (high) {
    case REGNORM:
        copy_region();
        save_var bufname = _cur_kill_buf;
        _is_rectangle = 0;
        _is_line = 0;
        break; 
    case REGRECT:
        process_block(COPY_BLOCK);
        break;
    case REGLINE:
        save_var point;
        save_var mark;
        line_up_mark_and_point();
        copy_region();
        save_var bufname = _cur_kill_buf;
        _is_rectangle = 0;
        _is_line = 1;
        break;
    default:
        highlight_on();
        break;
    }
  }

/*****************************************************************************
*                             kill-mark                                      *
*****************************************************************************/
command kill_mark()
  {
    int high;   

    if (in_rectangle_mode)
      {
        kill_region();
        return;
      }
    high = is_highlight_on();
    highlight_off();
    
    switch (high) {
    case REGNORM:
        if (mark == point)
            return;
        kill_region();
        save_var bufname = _cur_kill_buf;
        _is_rectangle = 0;
        _is_line = 0;
        break; 
    case REGRECT:
        process_block(KILL_BLOCK);
        break;
    case REGLINE:
        line_up_mark_and_point();
        kill_region();
        save_var bufname = _cur_kill_buf;
        _is_rectangle = 0;
        _is_line = 1;
        break;
    default:
        highlight_on();
        break;
    }
  }

command yank_mark()
  {
    int SaveCol;
    char *ThisBuff = bufname;
    short high;
    
    bufname = _cur_kill_buf;
    if (bufname == ThisBuff)
        return; 
    
    high = _is_rectangle ? REGRECT : _is_line ? REGLINE : REGNORM;
    bufname = ThisBuff;
    
    switch (high) {
    case REGNORM:
        yank();
        break; 
    case REGRECT:
        yank_block();
        break;
    case REGLINE:
        SaveCol = virtual_column();
        move_to_column(0);
        cursor_to_column = -1;
        yank();                            
        maybe_refresh();
        to_virtual_column(SaveCol);
        _oldto = point;
        break;
    }
  }

command yank_pop_mark()
  {
	if (!pop_retrieve())
		error("Must follow yank or yank-pop");
    yank_mark();
  }   

command copy_entire_line()    
  {
    int start, num, last = prev_cmd;
    int column = virtual_column();

    highlight_off();
    move_to_column(0);
    start = point; 
    if (iter < 0)
      {
        num = -iter;
        while (iter < 0)
          {
            iter++;
            if (!nl_reverse())
                break;
          }
        num += iter;
        move_to_column(0);
      }
    else
      {
        if (iter == 0)
            iter = 1;
        num = iter;
        while (iter > 0)
          {
            iter--;
            if (!nl_forward())
                break;
          }
        num -= iter;
      }                
    prev_cmd = last;
    do_save_only(start, point);
    to_virtual_column(column);

    save_var bufname = _cur_kill_buf;
    if (iter > 0)
        insert('\n');
    _is_rectangle = 0;
    _is_line = 1;

    if (num > 1)
        say("%d lines copied to %s", num, _cur_kill_buf);
    else
        say("Line copied to %s", _cur_kill_buf);
    iter = 0;
  }

command copy_line_or_mark()
  {
	if (has_arg || !is_highlight_on())
        copy_entire_line();
    else
        copy_mark();
  }

command kill_entire_line()    
  {
    int start, num, last = prev_cmd;
    int column = virtual_column();

    highlight_off();
    move_to_column(0);
    start = point; 
    if (iter < 0)
      {
        num = -iter;
        while (iter < 0)
          {
            iter++;
            if (!nl_reverse())
                break;
          }
        move_to_column(0);
        num += iter;
      }
    else
      {
        if (iter == 0)
            iter = 1;
        num = iter;
        while (iter > 0)
          {
            iter--;
            if (!nl_forward())
                break;
          }
        num -= iter;
      }                
    prev_cmd = last;
    do_save_kill(start, point);
    to_virtual_column(column);

    save_var bufname = _cur_kill_buf;
    if (iter > 0)
        insert('\n');
    _is_rectangle = 0;
    _is_line = 1;

    if (num > 1)
        say("%d lines copied to %s", num, _cur_kill_buf);
    else
        say("Line copied to %s", _cur_kill_buf);
    iter = 0;
  }

command kill_line_or_mark()
  {
	if (has_arg || !is_highlight_on())
        kill_entire_line();
    else
        kill_mark();
  }

command copy_entire_word()
    {
    save_var point;
    save_var mark;

    iter = 1;
    forward_word();
    mark = point;
    iter = 1;
    backward_word();
    copy_region();
    save_var bufname = _cur_kill_buf;
    _is_rectangle = 0;
    _is_line = 0;
    
    say("Word copied to %s", _cur_kill_buf);
    }

command kill_entire_word()
    {
    iter = 1;
    forward_word();
    mark = point;
    iter = 1;
    backward_word();
    kill_region();
    save_var bufname = _cur_kill_buf;
    _is_rectangle = 0;
    _is_line = 0;
    say("Word killed to %s", _cur_kill_buf);
    }

command set_mark_block()
    {
    mark = point;
    
    if (in_rectangle_mode)
        rectangle_mode();

    _is_rectangle = 1;
    _is_line = 0;
    highlight_on();
                         
    say("Block mark set.");
    }

command set_mark_lines()
    {
    save_var point;
        
    if (in_rectangle_mode)
        rectangle_mode();

    move_to_column(0);
    mark = point;

    _is_rectangle = 0;
    _is_line = 1;
    highlight_on();
    
    say("Lines mark set.");
    }                     
    
command set_mark_region()
    {
    mark = point;
                 
    if (in_rectangle_mode)
        rectangle_mode();
    
    _is_rectangle = 0;
    _is_line = 0;
    highlight_on();
                         
    say("Region mark set.");
    }

/*****************************************************************************
*                 copy, kill, or delete marked block                         *
*****************************************************************************/
process_block(OpCode)
    char OpCode;
    {
    int   i, j;
    int   BlockWidth;
    int   StartCol;
    int   EndCol;
    int   SavePoint;
    int   SaveMark;
    char  *ThisBuff  = bufname;

    block_setup(&StartCol, &EndCol);
    SavePoint = point;
    SaveMark  = mark;
    
    BlockWidth = EndCol - StartCol;
    if (BlockWidth == 0)
        return;
    move_to_column(StartCol);
    
    if (OpCode != COPY_BLOCK)
        SavePoint = point;
    
    if (OpCode != DELETE_BLOCK)
        {
        prev_cmd = 0;
        check_num_kill_buffers();
        push_kill();
		_kill_width[_num_kill] = BlockWidth;
        }
    
/*---------------------------------------------------------------------------*
*   transfer bytes from current buffer to current kill buffer                *
*---------------------------------------------------------------------------*/
    while (point <= mark)
        {
        i = point;
        move_to_column(EndCol);
        j = point - i;
        
        if (OpCode != DELETE_BLOCK)
            xfer(_cur_kill_buf, i, point);
        
        if (OpCode != COPY_BLOCK)
            delete(i, point);
        
        if (OpCode != DELETE_BLOCK)
            {
            bufname = _cur_kill_buf;
            i = point;
            point = size();
            if (character(point) == '\n')
                replace(point, ' ');
            if (j != BlockWidth)
                {
                while (j++ < BlockWidth)
                    insert(' ');
                point = size();
                }
            bufname = ThisBuff;
            }
        
        nl_forward();
        move_to_column(StartCol);
        }

    if (OpCode != DELETE_BLOCK)
        {
        bufname = _cur_kill_buf;
        modified = 0; 
        _is_line = 0;
        _is_rectangle = 1;
        bufname = ThisBuff;
        this_cmd = 0;
        }
    
/*---------------------------------------------------------------------------*
*   restore point and mark                                                   *
*---------------------------------------------------------------------------*/
    if (OpCode == COPY_BLOCK)
        {
        point = SavePoint;
        mark  = SaveMark;
        }
    else
        {
        point = SavePoint;
        mark  = SavePoint;
        }
    
    }

yank_block()
    {
    int   BlockWidth;
    int   StartCol;
    int   SavePoint = point;
    int   i;
    char  *ThisBuff = bufname;

	undo_flag = UNDO_FLAG_KILL;	/* mark in undo list */
/*---------------------------------------------------------------------------*
*   set mark to insert point                                                 *
*   get block width                                                          *
*---------------------------------------------------------------------------*/
    mark = point;
    StartCol = current_column();
    bufname = _cur_kill_buf;
    point = 0;
    BlockWidth = _kill_width[_num_kill];

/*---------------------------------------------------------------------------*
*   transfer characters from current kill buffer to current buffer           *
*---------------------------------------------------------------------------*/
    while (point < size())
        {
        i = point;
        point += BlockWidth;
        xfer(ThisBuff, i, point);
        
/*---------------------------------------------------------------------------*
*   increment insert point to next line in current buffer                    *
*   add a new line to end of buffer if necessary                             *
*---------------------------------------------------------------------------*/
        bufname = ThisBuff;
        if (!nl_forward())
            insert('\n');
        move_to_column(StartCol);
        if (current_column() < StartCol)
            to_column(StartCol);
        bufname = _cur_kill_buf;
        }
        
    bufname = ThisBuff;
    nl_reverse();
    move_to_column(StartCol + BlockWidth);
    mark = point;
    point = SavePoint;
    _oldfrom = point;
    _oldto = mark;
    }


/*****************************************************************************
*                     setup for block operations                             *
*****************************************************************************/
block_setup(StartCol, EndCol)
    int *StartCol;
    int *EndCol;
    {
        
    int i, j, k;
/*---------------------------------------------------------------------------*
*   get column number of opposite corners                                    *
*   swap mark and point if necessary to put point on one of upper corners    *
*---------------------------------------------------------------------------*/
    if (point > mark)
        {
        j = current_column();       /* one of the lower corners */
        k = point;
        point = mark;
        i = current_column();       /* one of the upper corners */
        mark = k;
        }
    else
        {
        i = current_column();       /* one of the upper corners */
        k = point;
        point = mark;
        j = current_column();       /* one of the lower corners */
        point = k;
        }
    
/*---------------------------------------------------------------------------*
*   compute block width                                                      *
*---------------------------------------------------------------------------*/
    if (j > i)
        {
        *StartCol = i;
        *EndCol = j;
        }
    else
        {
        *StartCol = j;
        *EndCol = i;
        }
    
    }

/*****************************************************************************
*                 replacement for push_kill() in kill.e                      *
*****************************************************************************/

new_mark_push_kill()
    {
    char *thisbuff = bufname;
    
    mark_push_kill();         /* old push_kill */
    bufname = _cur_kill_buf;
    _is_rectangle = 0;
    _is_line = 0;
    bufname = thisbuff;
    }


when_loading()
    {
/*---------------------------------------------------------------------------*
*   rename push_kill to mark_push_kill                                       *
*   rename new_mark_push_kill to push_kill                                   *
*---------------------------------------------------------------------------*/
    if (!find_index("mark_push_kill"))
        replace_name("push_kill", "mark_push_kill");
    else
        drop_name("push_kill");
    
    replace_name("new_mark_push_kill", "push_kill");
    
    }
